/*
 * XYZ Library - window.c
 *
 * Copyright (C) 2001 Woodley Packard
 * 
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/keysymdef.h>
#include	<X11/Xos.h>
#include	<sys/ipc.h>
#include	<sys/shm.h>

#include	<yz/window.h>
#include	<yz/draw.h>
#include	"window_data.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<time.h>

font_t		*font_from_name(char	*fname);

Display		*display = 0;
Colormap	cmap;
int			screen = 0;

window_t	*yz_current_window = 0;
window_data	*yz_d = 0;

int			__xyz_window_list_n = 0;
window_t	**__xyz_window_list = 0;

static void	register_xyz_window(window_t	*w)
{
	__xyz_window_list_n++;
	__xyz_window_list = realloc(__xyz_window_list, sizeof(window_t*) * __xyz_window_list_n);
	__xyz_window_list[__xyz_window_list_n-1] = w;
}

static void	deregister_xyz_window(window_t	*w)
{
	int	i;

	for(i=0;i<__xyz_window_list_n;i++)
		if(__xyz_window_list[i] == w)break;
	if(i>=__xyz_window_list_n)return;
	__xyz_window_list_n--;
	memmove(__xyz_window_list+i, __xyz_window_list+i+1, (__xyz_window_list_n-i)*sizeof(window_t*));
}

event_t	dequeue_window_event(window_t	*w);

int	ehand(Display	*dpy, XErrorEvent	*err)
{
	//printf("X-Error\n");
}

init_xyz()
{
	if(!(display=XOpenDisplay(0)))exit(-1);
	screen = DefaultScreen(display);
	cmap = DefaultColormap(display, screen);
	XSetErrorHandler(ehand);
#ifdef USE_PANGO
	pango_init();
#endif
}

Pixmap	make_x_pixmap(window_data	*d, int	width, int	height)
{
	Pixmap	pm;

	if(d->have_xshm)
		pm = XShmCreatePixmap(display, RootWindow(display, screen), d->buffer, &d->seginfo, width, height, DefaultDepth(display, screen));
	else pm = XCreatePixmap(display, RootWindow(display, screen), width, height, DefaultDepth(display, screen));
	return pm;
}

// MWM hints: copied from Lesstif by way of some other open source packages
typedef struct {
	unsigned long flags;
	unsigned long functions;
	unsigned long decorations;
	long input_mode;
	unsigned long status;
} MotifWmHints, MwmHints;

static window_t	*xyz_make_window(Window	xw, int	width, int	height, char	*name)
{
	window_data	*d;
	window_t	*yzw;
	static int	seqn = 0;
	int			shmlen;
	XGCValues	val;
	XTextProperty	name_tp;

	yzw = (window_t*)malloc(sizeof(window_t));
	d = (window_data*)malloc(sizeof(window_data));
	yzw->impl_data = d;
	yzw->width = width;
	yzw->height = height;
	yzw->wseq = ++seqn;
	yzw->pen = (pen_t*)malloc(sizeof(pen_t));
	yzw->pen->size = 1;
	yzw->pen->color.r = yzw->pen->color.g = yzw->pen->color.b = 0;
	yzw->font = 0;
	yzw->origin_x = yzw->origin_y = 0;
	d->col = BlackPixel(display, screen);
	d->have_custom_cursor = 0;
	d->xcursor = -1;
	d->xw = xw;
	d->cs.a = 0;
	d->cs.n = 0;
	d->cs.r = 0;
	d->os.a = 0;
	d->os.n = 0;
	d->os.p = 0;
	shmlen = (DefaultDepth(display, screen)+7)/8 * width * height;
	shmlen = (shmlen + 3)&(~3);	// align size to 4-byte boundary
	d->seginfo.shmseg = 0;
	d->seginfo.shmaddr = 0;
	d->seginfo.shmid = shmget(IPC_PRIVATE, shmlen, IPC_CREAT|0777);
	d->buffer = d->seginfo.shmaddr = shmat(d->seginfo.shmid, 0, 0);
	d->seginfo.readOnly = 0;
	//printf("before: shmaddr = %p, shmid = %p, shmseg = %p\n", d->seginfo.shmaddr,
				//d->seginfo.shmid, d->seginfo.shmseg);

	if(XShmQueryExtension(display))
	{
		if(!XShmAttach(display, &d->seginfo))
		{
			d->have_xshm = 1;
			//printf("after: shmaddr = %p, shmid = %p, shmseg = %p\n", d->seginfo.shmaddr,
						//d->seginfo.shmid, d->seginfo.shmseg);
		}
		else
		{
			d->have_xshm = 0;
			shmdt(d->seginfo.shmaddr);
			shmctl(d->seginfo.shmid, IPC_RMID, 0);
			d->seginfo.shmaddr = 0;
			//printf("no x-shared-memory availible\n");
		}
	}
	//else printf("server doesn't support XShm extension\n");

	Xutf8SetWMProperties(display, xw, name, name, 0, 0, 0, 0, 0);

	d->wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", 0);
	XSetWMProtocols(display, d->xw, &d->wm_delete_window, 1);

	d->xp = make_x_pixmap(d, width, height);

	XMapWindow(display, d->xw);
	val.function = GXcopy;
	d->gc = XCreateGC(display, d->xw, GCFunction, &val);
	val.function = GXcopy;
	d->pgc = XCreateGC(display, d->xp, GCFunction, &val);
	
	// clear the pixmap, since X doesn't do this for us...
		
		XSetForeground(display, d->pgc, WhitePixel(display, screen));
		XFillRectangle(display, d->xp, d->pgc, 0,0, yzw->width, yzw->height);
		XSetForeground(display, d->pgc, BlackPixel(display, screen));

	d->x = d->y = 0; d->b = 0;
	yzw->font = font_from_name("system");
#ifdef USE_PANGO
	d->fnt = ((struct font_data*)yzw->font->impl_data)->fd;
#else
	d->fnt = ((struct font_data*)yzw->font->impl_data)->xfs;
#endif
	d->evq = 0; d->evqd = 0;
	d->evf = 0; d->wd = 0;
	d->buffer_mode = YZ_BUFFER_BOTH;
	d->wd = 0;	d->is_postscript = 0;
	//XSetFont(display, d->gc, d->fnt->fid);
	//XSetFont(display, d->pgc, d->fnt->fid);
	strncpy(yzw->name, name, 64);
	yzw->name[63] = 0;
	usleep(10000);

	register_xyz_window(yzw);
	return yzw;
}

window_t	*yzNewWindowFlagsAt(char	*name, int	x, int	y, int	width, int	height, int	flags)
{
	Window	xw;
	window_t	*yzw;
	XSetWindowAttributes xa;
	MwmHints	mwh = { 2, 0, 1, 0, 0 };

	xw = XCreateSimpleWindow(display, RootWindow(display, screen),
				x, y, width, height, 1, 1, WhitePixel(display, screen));
	if(flags & 32)
	{
		// borderless
		mwh.decorations = 0;
		xa.event_mask = XYZ_EVM & ~ EnterWindowMask & ~ LeaveWindowMask;
		xa.override_redirect = 1;	// look out! muahahaha!
		XChangeWindowAttributes(display, xw, CWEventMask | CWOverrideRedirect, &xa);
	}
	else
	{
		xa.event_mask = XYZ_EVM;
		xa.override_redirect = 0;	// don't use OvRed for borderful windows
		XChangeWindowAttributes(display, xw, CWEventMask, &xa);
	}
	if(flags & 64)
	{
		// transient/model
		if(!yz_current_window)
		{
			fprintf(stderr, "Tried to create a transient with no parent, don't know how!\n");
			exit(-1);
		}
		XSetTransientForHint(display, xw, yz_d->xw);
	}
	else XSetTransientForHint(display, xw, 0);
	XChangeProperty(display, xw, XInternAtom(display, "_MOTIF_WM_HINTS",0),
								XInternAtom(display, "_MOTIF_WM_HINTS",0), 32,
								PropModeReplace, (char*)&mwh, 5);
	yzw = xyz_make_window(xw, width, height, name);
	if(flags & 32)
	{
		// borderless
		XGrabPointer(display, xw, 0, PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
					GrabModeAsync, GrabModeAsync, 0, None, CurrentTime);
	}
	XFlush(display);
	return yzw;
}

window_t	*yzNewWindow(char	*name, unsigned short	width, unsigned short	height)
{
	Window	xw;
	window_t	*yzw;
	XSetWindowAttributes xa;
	MwmHints	mwh = { 2, 0, 1, 0, 0 };

	if(!display)init_xyz();
	xw = XCreateSimpleWindow(display, RootWindow(display, screen),
				0, 0, width, height, 1, 1, WhitePixel(display, screen));
	xa.event_mask = XYZ_EVM;
	XChangeWindowAttributes(display, xw, CWEventMask, &xa);
	XChangeProperty(display, xw, XInternAtom(display, "_MOTIF_WM_HINTS",0),
								XInternAtom(display, "_MOTIF_WM_HINTS",0), 32,
								PropModeReplace, (char*)&mwh, 5);
	yzw = xyz_make_window(xw, width, height, name);
	//XSelectInput(display, xw, StructureNotifyMask);
	XFlush(display);
	return yzw;
}

window_t	*yzFullScreen()
{
	window_t	*yzw;
	Window	xw;
	MwmHints	mwh = { 2, 0, 0, 0, 0 };

	if(!display)init_xyz();
	xw = RootWindow(display, screen);
	XChangeProperty(display, xw, XInternAtom(display, "_MOTIF_WM_HINTS",0),
								XInternAtom(display, "_MOTIF_WM_HINTS",0), 32,
								PropModeReplace, (char*)&mwh, 5);
	yzw = xyz_make_window(xw, XDisplayWidth(display, screen),
								  XDisplayHeight(display, screen), "FullScreen");
	XFlush(display);
	return yzw;
}

extern int	yz_fd;
extern FILE	*ps_f;

yzDeleteWindow(window_t	*w)
{
	window_data	*wd;
	window_t	*sw;
	event_t		ev;

	if(!w)return -1;
	deregister_xyz_window(w);
	if(w->font && ((struct font_data*)w->font->impl_data)->auto_dealloc)free_font(w->font);
	w->font = 0;
	wd = w->impl_data;
	XUngrabPointer(display, CurrentTime);
	XUnmapWindow(display, wd->xw);
	XFreeGC(display, wd->pgc);
	XFreeGC(display, wd->gc);
	XFreePixmap(display, wd->xp);
	if(wd->seginfo.shmaddr)
	{
		XShmDetach(display, &wd->seginfo);
		shmdt(wd->seginfo.shmaddr);
		shmctl(wd->seginfo.shmid, IPC_RMID, 0);
	}
	XDestroyWindow(display, wd->xw);
	XFlush(display);
	sw = yz_current_window;
	yz_current_window = w;
	yz_d = yz_current_window->impl_data;
	while(wd->evf)yzRemoveEventFilter(wd->evf->filter);
	while(wd->wd)yzPutWindowData(wd->wd->key, 0);
	do { ev = dequeue_window_event(w); } while (ev.type != YZ_IDLE);
	yz_current_window = (sw==w)?0:sw;
	if(sw==w)
	{
		if(ps_f)fclose(ps_f);
		ps_f = 0;
	}
	free((char*)w);
	free((char*)wd);
	if(yz_current_window)yz_d = yz_current_window->impl_data;
	else yz_d = 0;
	return 0;
}

setup_ps_fptrs()
{
	extern int	(*yzLine2i2i)(int,int,int,int);
	extern int	(*yzRect2i2i)(int,int,int,int);
	extern int	(*yzPenColor)(int, int, int);
	extern int	(*yzCircle3i)(int, int, int);
	extern int	(*yzText2i1p1i)(int,int,char*,int);
	extern int	(*yzStringSize)(char*,int);
	extern int	(*yzDefaultLineHeight)();
	extern int	yzpsLine2i2i(int,int,int,int);
	extern int	yzpsRect2i2i(int,int,int,int);
	extern int	yzpsPenColor3i(int,int,int);
	extern int	yzpsCircle3i(int, int, int);
	extern int	yzpsText2i1p1i(int,int,char*,int);
	extern int	yzpsStringSize(char*,int);
	extern int	yzpsDefaultLineHeight();

	yzLine2i2i = yzpsLine2i2i;
	yzRect2i2i = yzpsRect2i2i;
	yzPenColor3i = yzpsPenColor3i;
	yzCircle3i = yzpsCircle3i;
	yzText2i1p1i = yzpsText2i1p1i;
	yzStringSize = yzpsStringSize;
	yzDefaultLineHeight = yzpsDefaultLineHeight;
	yz_fd = yz_d->fd;
	ps_f = fdopen(yz_fd, "w");
}

setup_x11_fptrs()
{
	extern int	(*yzLine2i2i)(int,int,int,int);
	extern int	(*yzRect2i2i)(int,int,int,int);
	extern int	(*yzPenColor)(int, int, int);
	extern int	(*yzCircle3i)(int, int, int);
	extern int	(*yzText2i1p1i)(int,int,char*,int);
	extern int	(*yzStringSize)(char*,int);
	extern int	(*yzDefaultLineHeight)();
	extern int	yzwLine2i2i(int,int,int,int);
	extern int	yzwRect2i2i(int,int,int,int);
	extern int	yzwPenColor3i(int,int,int);
	extern int	yzwCircle3i(int, int, int);
	extern int	yzwText2i1p1i(int,int,char*,int);
	extern int	yzwStringSize(char*,int);
	extern int	yzwDefaultLineHeight();

	yzLine2i2i = yzwLine2i2i;
	yzRect2i2i = yzwRect2i2i;
	yzPenColor3i = yzwPenColor3i;
	yzCircle3i = yzwCircle3i;
	yzText2i1p1i = yzwText2i1p1i;
	yzStringSize = yzwStringSize;
	yzDefaultLineHeight = yzwDefaultLineHeight;
	yz_fd = -1;
	if(ps_f)
	{
		fclose(ps_f);
		ps_f = 0;
	}
}

window_t	*yzSelectWindow(window_t	*w)
{
	window_t	*ow;
	
	if(!w)return;
	//fprintf(stderr, "Selecting window %s\n", w->name);
	ow = yz_current_window;
	yz_current_window = w;
	yz_d = w->impl_data;
	if(yz_d->is_postscript)setup_ps_fptrs();
	else setup_x11_fptrs();
	return ow;
}

window_t	*yzGetSelectedWindow(window_t	*w)
{
	return yz_current_window;
}

yzGetWindowCorner(int	*cx, int	*cy)
{
	//Window	root;
	//unsigned int	wid, hei, bwid, dep;
	//XWindowAttributes	attr;
	int	winx, winy, rootx, rooty;
	Window	root, child;
	unsigned int	mask;

	if(!yz_current_window)
	{
		*cx = 100;
		*cy = 100;
		return -1;
	}
	//XGetGeometry(display, yz_d->xw, &root, cx, cy, &wid, &hei, &bwid, &dep);
	//printf("getting coords for wd %p, xw %d\n", yz_d, yz_d->xw);
	//XGetWindowAttributes(display, yz_d->xw, &attr);
	//*cx = attr.x;
	//*cy = attr.y;
	if(XQueryPointer(display, yz_d->xw, &root, &child,
						&rootx, &rooty, &winx, &winy, &mask))
	{
		*cx = rootx-winx;
		*cy = rooty-winy;
	}
	else
	{
		*cx = 100;
		*cy = 100;
	}
	return 0;
}

int	yzPutWindowData(char	*key, void	*data)
{
	window_data_item	*wd, *s;

	if(!yz_current_window)return -1;
	s = 0;
	wd = yz_d->wd;
	while(wd)
	{
		if(!strcmp(wd->key, key))
		{
			// it's already there
			if(data)wd->data = data;
			else
			{
				// we have to remove it (data = 0 means we don't want it)
				if(wd->next)wd->next->prev = wd->prev;
				if(wd->prev)wd->prev->next = wd->next;
				if(yz_d->wd == wd)yz_d->wd = wd->next;
				free((char*)wd);
				return 0;
			}
		}
		s = wd;
		wd = wd->next;
	}
	if(!wd)
	{
		// we need to add a new item
		if(data) // don't bother adding nulls...
		{
			wd = (window_data_item*)malloc(sizeof(window_data_item));
			strcpy(wd->key, key); wd->data = data;
			wd->next = 0; wd->prev = s;
			if(s)s->next = wd;
			else yz_d->wd = wd;
		}
	}
	return 0;
}

void	*yzGetWindowData(char	*key)
{
	window_data_item	*d;
	
	if(!yz_current_window)return 0;
	d = yz_d->wd;
	while(d)
	{
		if(!strcmp(d->key, key))return d->data;
		d = d->next;
	}
	return 0;
}

int	yzUpdateBuffer(int	mode)
{
	if(!yz_current_window)return -1;
	if(mode == YZ_BUFFER_BACK)
		XCopyArea(display, yz_d->xw, yz_d->xp, yz_d->pgc, 0, 0,
			yz_current_window->width, yz_current_window->height, 0, 0);
	if(mode == YZ_BUFFER_FRONT)
		XCopyArea(display, yz_d->xp, yz_d->xw, yz_d->gc, 0, 0,
			yz_current_window->width, yz_current_window->height, 0, 0);
	XFlush(display);
	XSync(display, 0);
	return 0;
}

int	yzBufferMode(int	mode)
{
	if(!yz_current_window)return -1;
	if(mode != YZ_BUFFER_BOTH && mode != YZ_BUFFER_FRONT && mode != YZ_BUFFER_BACK)return -1;
	yz_d->buffer_mode = mode;
	do_select_font();
	do_select_pen();
	return 0;
}
